
#include <clickhouse/client.h>
#include <fmt/chrono.h>
#include <fmt/format.h>
#include <fmt/std.h>
#include <zstd.h>
// #include <libexplain/ferror.h>
#include <cstddef>
#include <cstdint>
#include <exception>
#include <filesystem>
#include <nlohmann/json.hpp>
#include <nlohmann/json_fwd.hpp>

#include "tsdb_hf.h"

/// @brief Custom C++ wrapper for ZSTD.
namespace zstd_cpp {

struct zstd_error : std::runtime_error {
    zstd_error(const std::string& msg) : std::runtime_error(msg) {}
    zstd_error(const char* msg) : std::runtime_error(msg) {}
};

void check_zstd(size_t code, std::string_view msg = "ZSTD error") {
    if (ZSTD_isError(code)) {
        throw zstd_error(fmt::format("{}: {}", msg, ZSTD_getErrorName(code)));
    }
}

struct CCtx {
    CCtx() : cctx_(ZSTD_createCCtx()) {
        if (!cctx_) {
            throw zstd_error("Failed to create ZSTD_CCtx");
        }
    }
    CCtx(std::nullptr_t) : cctx_(nullptr) {}
    ~CCtx() { ZSTD_freeCCtx(cctx_); }
    CCtx(const CCtx&) = delete;
    CCtx(CCtx&& other) : CCtx(nullptr) { std::swap(cctx_, other.cctx_); }
    CCtx& operator=(CCtx other) {
        std::swap(cctx_, other.cctx_);
        return *this;
    }

    void reset() { ZSTD_CCtx_reset(cctx_, ZSTD_reset_session_only); }

    void set_compression_level(int compressionLevel) {
        check_zstd(ZSTD_CCtx_setParameter(cctx_, ZSTD_c_compressionLevel, compressionLevel));
    }

    /// @brief Performs a streaming compression operation.
    /// @param output
    /// @param input
    /// @param endOp
    /// @return The approximate remaining bytes to flush to the output buffer.
    size_t compress_stream(ZSTD_outBuffer& output, ZSTD_inBuffer& input, ZSTD_EndDirective endOp) {
        size_t remaining;
        remaining = ZSTD_compressStream2(cctx_, &output, &input, endOp);
        check_zstd(remaining);
        return remaining;
    }

    /// @brief Performs a simple streaming compression operation. This is a convenience function
    /// to simplify handling of compressed data.
    /// @param src The source data. Specify `nullptr` to end the current compression session.
    /// @param srcSize The size of the source data.
    /// @param dst The destination buffer. Only used as a temporary buffer.
    /// @param dstCapacity The capacity of the destination buffer.
    /// @param callback The callback function that will be called with the compressed data.
    /// Should have the signature `fn(const void* compressed_data, size_t size) -> void`.
    template <typename F>
    void compress_stream_simple(const void* src, size_t srcSize, void* dst, size_t dstCapacity, F&& callback) {
        ZSTD_inBuffer input   = {src, srcSize, 0};
        ZSTD_outBuffer output = {dst, dstCapacity, 0};

        if (src == nullptr) {
            input.size = 0;
            size_t remaining;
            // Must flush everything inside the internal buffer now.
            do {
                output.pos = 0;
                remaining  = compress_stream(output, input, ZSTD_e_end);
                callback(output.dst, output.pos);
            } while (remaining > 0);
            return;
        }

        // It's okay to stop as soon as we've drained the input buffer, and leave some
        // remaining compressed data in the ZSTD's internal buffer.
        while (input.pos < input.size) {
            output.pos = 0;
            compress_stream(output, input, ZSTD_e_continue);
            callback(output.dst, output.pos);
        }
    }

private:
    ZSTD_CCtx* cctx_;
};

}  // namespace zstd_cpp

namespace tsdb_hf_cpp {

void throw_errno(std::string_view msg = "errno") {
    throw std::runtime_error(fmt::format("{}: {}", msg, strerror(errno)));
}

void check_errno(bool condition, std::string_view msg = "errno") {
    if (!condition) {
        throw_errno(msg);
    }
}

std::filesystem::path concat_path(std::string_view base, std::string_view name) {
    return std::filesystem::path(base) / name;
}

/// @brief Renames a file to a new extension. If the extension is empty,
/// the file will be renamed to have no extension.
void rename_to_extension(const std::filesystem::path& path, const std::filesystem::path& extension = std::filesystem::path()) {
    auto new_path = path;
    new_path.replace_extension(extension);
    std::filesystem::rename(path, new_path);
}

/// @brief RAII wrapper for a binary file. Better than std::fstream.
struct bin_file {
    bin_file() : file_(nullptr) {}
    bin_file(std::nullptr_t) : file_(nullptr) {}
    bin_file(const std::filesystem::path& path, const char* mode) {
        file_ = fopen(path.c_str(), mode);
        if (!file_) {
            throw_errno(fmt::format("Failed to open file `{}` with mode {}", path, mode));
        }
    }
    ~bin_file() {
        if (file_) {
            // Failures are ignored here
            fclose(file_);
        }
    }
    bin_file(const bin_file&) = delete;
    bin_file(bin_file&& other) : file_(other.file_) { other.file_ = nullptr; }
    bin_file& operator=(bin_file other) {
        std::swap(file_, other.file_);
        return *this;
    }

    operator bool() const { return file_ != nullptr; }

    void close() {
        if (auto file = std::exchange(file_, nullptr)) {
            check_errno(fclose(file) == 0, "Failed to close file");
        }
    }
    bool is_open() const { return file_ != nullptr; }

    size_t tell() const {
        auto pos = ftell(file_);
        check_errno(pos != -1, "Failed to get file position");
        return pos;
    }
    void seek(ssize_t pos = 0, int whence = SEEK_SET) { check_errno(fseek(file_, pos, whence) == 0, "Failed to seek in file"); }
    void rewind() { check_errno(fseek(file_, 0, SEEK_SET) == 0, "Failed to rewind file"); }
    size_t read(void* data, size_t size) {
        auto count = fread(data, 1, size, file_);
        if (ferror(file_)) {
            throw_errno("Failed to read from file");
        }
        return count;
    }
    void read_exact(void* data, size_t size) { check_errno(fread(data, 1, size, file_) == size, "Failed to read from file"); }
    void write(const void* data, size_t size) { check_errno(fwrite(data, 1, size, file_) == size, "Failed to write to file"); }
    void flush() { check_errno(fflush(file_) == 0, "Failed to flush file"); }

private:
    FILE* file_;
};

// Forward declaration
void tsdb_entry_notify_stream_close(tsdb_entry::impl* entry);
size_t tsdb_entry_get_part_size(tsdb_entry::impl* entry) noexcept;

struct StreamMetadata {
    std::string name;
    uint64_t timestampOffset;
    uint64_t interval;
};

struct tsdb_entry_stream::impl {
    impl(tsdb_entry::impl* parent, std::filesystem::path stream_base_dir, StreamMetadata metadata)
        : parent_(parent),
          stream_base_dir_(std::move(stream_base_dir)),
          metadata_(std::move(metadata)),
          part_max_size_(tsdb_entry_get_part_size(parent_)) {
        // TODO: Maybe recycle cctx_?
        cctx_.reset();

        rewrite_metadata();
    }
    ~impl() {
        // Notify the parent that the stream is closed
        tsdb_entry_notify_stream_close(parent_);
    }

    // WARN: You must call this method exactly once before dropping the object.
    void close() {
        // Flush the remaining data
        close_current_part();

        // Emit the metadata about the stream and the data parts
        rewrite_metadata();
    }

    void insert_point(double value) { insert_points(&value, 1); }
    void insert_points(const double* values, size_t count) {
        while (count > 0) {
            if (cur_part_remaining_size_ == 0) {
                open_new_part();
            }

            auto to_write = std::min(count, cur_part_remaining_size_);
            cctx_.compress_stream_simple(values, to_write * sizeof(double), buffer_, sizeof buffer_,
                                         [this](const void* data, size_t size) { write_to_disk(data, size); });
            cur_part_remaining_size_ -= to_write;
            cur_points_count_ += to_write;
            values += to_write;
            count -= to_write;
        }
    }

private:
    nlohmann::json dump_metadata() {
        return {
            {"name", metadata_.name},
            {"timestamp_offset", metadata_.timestampOffset},
            {"interval", metadata_.interval},
            {"parts_count", cur_parts_index_},
            {"points_per_part", part_max_size_},
            {"total_points", cur_points_count_},
        };
    }

    void rewrite_metadata() {
        auto metadata      = dump_metadata().dump(4);
        auto metadata_path = stream_base_dir_ / "metadata.json.part";
        auto metadata_file = bin_file(metadata_path, "wb");
        metadata_file.write(metadata.data(), metadata.size());
        metadata_file.close();
        rename_to_extension(metadata_path);
    }

    void write_to_disk(const void* data, size_t size) {
        cur_part_file_.write(data, size);

        // TODO: Maybe flush eagerly?
        // if (cur_part_file_) {
        //     cur_part_file_.flush();
        // }
        // if (cur_part_remaining_size == 0) {
        //     close_current_part();
        // }
    }

    void open_new_part() {
        close_current_part();

        auto path = stream_base_dir_ / fmt::format("part_{:09d}.zst.part", cur_parts_index_);
        if (std::filesystem::exists(path)) {
            throw std::runtime_error(fmt::format("Part file already exists: {}", path));
        }
        cur_part_file_           = bin_file(path, "wb");
        cur_part_file_path_      = std::move(path);
        cur_part_remaining_size_ = part_max_size_;
        cur_parts_index_++;
    }

    void close_current_part() {
        if (cur_part_file_) {
            // Flush the remaining data
            cctx_.compress_stream_simple(nullptr, 0, buffer_, sizeof(buffer_),
                                         [this](const void* data, size_t size) { write_to_disk(data, size); });

            // Commit the previous part (remove the `.part` extension)
            cur_part_file_.close();
            rename_to_extension(cur_part_file_path_);
        }
    }

    tsdb_entry::impl* const parent_;
    StreamMetadata metadata_;
    zstd_cpp::CCtx cctx_;
    std::filesystem::path stream_base_dir_;
    std::filesystem::path cur_part_file_path_;
    bin_file cur_part_file_;
    // NOTE: The size represents the remaining points count, not the remaining bytes.
    size_t cur_parts_index_         = 0;
    size_t cur_part_remaining_size_ = 0;
    const size_t part_max_size_     = -1;
    size_t cur_points_count_        = 0;
    char buffer_[4 * 1024];
};

tsdb_entry_stream::tsdb_entry_stream(std::unique_ptr<impl> impl) : impl_(std::move(impl)) {
}
tsdb_entry_stream::~tsdb_entry_stream() {
    try {
        close();
    } catch (const std::exception& e) {
        fprintf(stderr, "Unhandled exception in ~tsdb_entry_stream: %s\n", e.what());
    }
}
tsdb_entry_stream::tsdb_entry_stream(tsdb_entry_stream&& other) : impl_(std::move(other.impl_)) {
}
void tsdb_entry_stream::close() {
    if (auto impl = std::move(impl_)) {
        impl->close();
    }
}
void tsdb_entry_stream::insert_point(double value) {
    impl_->insert_point(value);
}
void tsdb_entry_stream::insert_points(const double* values, size_t count) {
    impl_->insert_points(values, count);
}

struct tsdb_entry::impl {
    ~impl() {}

    // WARN: You must call this method exactly once before dropping the object.
    void close() {
        if (open_streams_count_ > 0) {
            // TODO: Maybe fail-fast?
            // Fail-fast
            throw std::runtime_error("Cannot close tsdb_entry while there are open streams");
        }
    }

    void set_data_dir(std::string_view data_dir) {
        // TODO: Maybe sanitize the path?
        data_dir_ = data_dir;
    }

    void set_part_size(size_t size) {
        if (size < 1) {
            throw std::invalid_argument("Part size must be positive");
        }
        part_size_ = size;
    }

    tsdb_entry_stream new_stream(std::string name, uint64_t timestampOffset, uint64_t interval) {
        // // Ensure base dir
        // std::filesystem::create_directories(data_dir_);

        auto dir_name        = fmt::format("{}-{:%Y-%m-%d_%H-%M-%S}", name, std::chrono::system_clock::now());
        auto stream_base_dir = concat_path(data_dir_, dir_name);
        auto stream_metadata = StreamMetadata{std::move(name), timestampOffset, interval};
        std::filesystem::create_directories(stream_base_dir);
        auto stream = tsdb_entry_stream(std::make_unique<tsdb_entry_stream::impl>(this, stream_base_dir, std::move(stream_metadata)));
        open_streams_count_++;
        return stream;
    }

    friend void tsdb_entry_notify_stream_close(tsdb_entry::impl* entry) {
        if (entry->open_streams_count_ == 0) {
            // Fail-fast
            throw std::runtime_error("open_streams_count_ is already 0");
        }
        entry->open_streams_count_--;
    }

    friend size_t tsdb_entry_get_part_size(tsdb_entry::impl* entry) noexcept { return entry->part_size_; }

private:
    std::string data_dir_      = "data";
    size_t part_size_          = 10000;
    size_t open_streams_count_ = 0;
};

tsdb_entry::tsdb_entry() : impl_(std::make_unique<impl>()) {
}
tsdb_entry::~tsdb_entry() {
    close();
}
tsdb_entry::tsdb_entry(tsdb_entry&& other) : impl_(std::move(other.impl_)) {
}
void tsdb_entry::close() {
    if (auto impl = std::move(impl_)) {
        impl->close();
    }
}
void tsdb_entry::set_data_dir(std::string_view data_dir) {
    impl_->set_data_dir(data_dir);
}
void tsdb_entry::set_part_size(size_t size) {
    impl_->set_part_size(size);
}
tsdb_entry_stream tsdb_entry::new_stream(std::string name, uint64_t timestampOffset, uint64_t interval) {
    return impl_->new_stream(name, timestampOffset, interval);
}

}  // namespace tsdb_hf_cpp
